﻿#coding=UTF-8
#enhancedListViewSupport.py
#A Global Plugin adding enhanced list view support to NVDA
#Written by Peter Vagner <peter.v@datagate.sk>

import globalPluginHandler
from NVDAObjects.IAccessible.sysListView32 import * #List, ListItem
import ui
import scriptHandler
import winUser
import winKernel
from NVDAObjects import NVDAObject
from NVDAObjects.IAccessible import IAccessible
import oleacc
#from logHandler import log
import textInfos
import watchdog
import languageHandler
import gettext
import os
import globalVars

try:
	_ = gettext.translation(__file__.split('\\')[-1].split('.')[:-1][0], '%s\\locale\\globalPlugins'%os.path.abspath(globalVars.appArgs.configPath),[languageHandler.curLang]).ugettext
except IOError:
	pass


 #Window messages
LVM_GETITEMTEXTW=LVM_FIRST+115
LVM_GETHEADER=LVM_FIRST+31
LVM_GETCOLUMNORDERARRAY=LVM_FIRST+59
LVM_GETCOLUMNW=LVM_FIRST+95
LVM_GETSELECTEDCOUNT =(LVM_FIRST+50)
LVNI_SELECTED =2
LVM_GETNEXTITEM =(LVM_FIRST+12)


#column mask flags
LVCF_FMT=1
LVCF_WIDTH=2
LVCF_TEXT=4
LVCF_SUBITEM=8
LVCF_IMAGE=16
LVCF_ORDER=32

CBEMAXSTRLEN=260

# listview header window messages
HDM_FIRST=0x1200
HDM_GETITEMCOUNT=HDM_FIRST


class LVItemStruct(Structure):
	_fields_=[
		('mask',c_uint),
		('iItem',c_int),
		('iSubItem',c_int),
		('state',c_uint),
		('stateMask',c_uint),
		('pszText',c_void_p),
		('cchTextMax',c_int),
		('iImage',c_int),
		('lParam',LPARAM),
		('iIndent',c_int),
		('iGroupID',c_int),
		('cColumns',c_uint),
		('puColumns',c_uint),
		('piColFmt',POINTER(c_int)),
		('iGroup',c_int),
	]

class LVCOLUMN(Structure):
	_fields_=[
		('mask',c_uint),
		('fmt',c_int),
		('cx',c_int),
		('pszText',c_void_p),
		('cchTextMax',c_int),
		('iSubItem',c_int),
		('iImage',c_int),
		('iOrder',c_int),
		('cxMin',c_int),
		('cxDefault',c_int),
		('cxIdeal',c_int),
	]

class EnhancedList(List):

	def _get_rowCount(self):
		return len(self.children)

	def _get_columnCount(self):
		headerHwnd= watchdog.cancellableSendMessage(self.windowHandle,LVM_GETHEADER,0,0)
		count=watchdog.cancellableSendMessage(headerHwnd,HDM_GETITEMCOUNT,0,0)
		if not count:
			count=1
		return count

	def _get_columnOrderArray(self):
		coa=(c_int *self.columnCount)()
		(processID,threadID)=winUser.getWindowThreadProcessID(self.windowHandle)
		processHandle=self.processHandle
		internalCoa=winKernel.virtualAllocEx(processHandle,None,sizeof(coa),winKernel.MEM_COMMIT,winKernel.PAGE_READWRITE)
		winKernel.writeProcessMemory(processHandle,internalCoa,byref(coa),sizeof(coa),None)
		res = watchdog.cancellableSendMessage(self.windowHandle,LVM_GETCOLUMNORDERARRAY, self.columnCount, internalCoa)
		if res:
			winKernel.readProcessMemory(processHandle,internalCoa,byref(coa),sizeof(coa),None)
		winKernel.virtualFreeEx(processHandle,internalCoa,0,winKernel.MEM_RELEASE)
		return coa

	def _get_selectedRows(self):
		numberOfRows=watchdog.cancellableSendMessage(self.windowHandle,LVM_GETSELECTEDCOUNT,0,0)
		if numberOfRows <=0:
			return None
		IAccessibles=[]
		index=-1
		index=winUser.sendMessage(self.windowHandle,LVM_GETNEXTITEM,index,LVNI_SELECTED)
		while index >-1:
			# childID 0 is the list view it-self, items are starting with childID 1
			childID=index+1
			IAccessibles.append(IAccessible(windowHandle=self.windowHandle,IAccessibleObject=self.IAccessibleObject,IAccessibleChildID=childID,event_windowHandle=self.event_windowHandle,event_objectID=self.event_objectID,event_childID=childID))
			index=winUser.sendMessage(self.windowHandle,LVM_GETNEXTITEM,index,LVNI_SELECTED)
		return IAccessibles


class EnhancedListItem(ListItem):
 
	def getcolumnText(self,index):
		buffer=None
		(processID,threadID)=winUser.getWindowThreadProcessID(self.windowHandle)
		processHandle=self.processHandle
		internalItem=winKernel.virtualAllocEx(processHandle,None,sizeof(LVItemStruct),winKernel.MEM_COMMIT,winKernel.PAGE_READWRITE)
		internalText=winKernel.virtualAllocEx(processHandle,None,CBEMAXSTRLEN*2,winKernel.MEM_COMMIT,winKernel.PAGE_READWRITE)
		item=LVItemStruct(iItem=self.IAccessibleChildID-1,mask=LVIF_TEXT|LVIF_COLUMNS,iSubItem=index,pszText=internalText,cchTextMax=CBEMAXSTRLEN)
		winKernel.writeProcessMemory(processHandle,internalItem,byref(item),sizeof(LVItemStruct),None)
		len = watchdog.cancellableSendMessage(self.windowHandle,LVM_GETITEMTEXTW, (self.IAccessibleChildID-1), internalItem)
		if len:
			winKernel.readProcessMemory(processHandle,internalItem,byref(item),sizeof(LVItemStruct),None)
			buffer=create_unicode_buffer(len)
			winKernel.readProcessMemory(processHandle,item.pszText,buffer,sizeof(buffer),None)
		winKernel.virtualFreeEx(processHandle,internalText,0,winKernel.MEM_RELEASE)
		winKernel.virtualFreeEx(processHandle,internalItem,0,winKernel.MEM_RELEASE)
		return buffer.value if buffer else None

	def getcolumnHeaderText(self,index):
		buffer=None
		(processID,threadID)=winUser.getWindowThreadProcessID(self.windowHandle)
		processHandle=self.processHandle
		internalColumn=winKernel.virtualAllocEx(processHandle,None,sizeof(LVCOLUMN),winKernel.MEM_COMMIT,winKernel.PAGE_READWRITE)
		internalText=winKernel.virtualAllocEx(processHandle,None,CBEMAXSTRLEN*2,winKernel.MEM_COMMIT,winKernel.PAGE_READWRITE)
		column=LVCOLUMN(mask=LVCF_TEXT,iSubItem=index,pszText=internalText,cchTextMax=CBEMAXSTRLEN)
		winKernel.writeProcessMemory(processHandle,internalColumn,byref(column),sizeof(LVCOLUMN),None)
		res = watchdog.cancellableSendMessage(self.windowHandle,LVM_GETCOLUMNW, index, internalColumn)
		if res:
			winKernel.readProcessMemory(processHandle,internalColumn,byref(column),sizeof(LVCOLUMN),None)
			buffer=create_unicode_buffer(column.cchTextMax)
			winKernel.readProcessMemory(processHandle,column.pszText,buffer,sizeof(buffer),None)
		winKernel.virtualFreeEx(processHandle,internalText,0,winKernel.MEM_RELEASE)
		winKernel.virtualFreeEx(processHandle,internalColumn,0,winKernel.MEM_RELEASE)
		return buffer.value if buffer else None

	def _get_lvName(self):
		coa=self.parent.columnOrderArray
		nameList=[]
		for c in coa:
#			column=self.getcolumnHeaderText(c)
			subItem=self.getcolumnText(c)
			colList=[]
#			if column:
#				colList.append(column)
			if subItem:
				colList.append(subItem)
			if len(colList):
				nameList.append(u": ".join(colList))
		return u"; ".join(nameList)

	def _get_value(self):
		if self.name:
			return None
		return super(EnhancedListItem,self).value

	def _get_name(self):
		name=self.lvName
		if (not name or name.isspace()) and self.windowStyle & LVS_OWNERDRAWFIXED:
			name=self.displayText
		if not name:
			return None
		#Some list view items in Vista can contain annoying left-to-right and right-to-left indicator characters which really should not be there.
		name=name.replace(u'\u200E','')
		name=name.replace(u'\u200F','')
		return name

	def formatColumn(self,index):
		coa=self.parent.columnOrderArray
		columnHeader=self.getcolumnHeaderText(coa[index])
		column=self.getcolumnText(coa[index])
		colList=[]
		if columnHeader:
			colList.append(columnHeader)
		if column:
			colList.append(column)
		if len(colList):
			text=u": ".join(colList)
		return text


class MozillaTable(IAccessible):

	def _get_selectedRows(self):
		if not hasattr(self,'IAccessibleTableObject'):
			return None
		numberOfRows=self.IAccessibleTableObject.nSelectedRows
		if numberOfRows <=0:
			return None
		rows=self.IAccessibleTableObject.selectedRows(numberOfRows)
		IAccessibles=[]
		for number in range(numberOfRows):
			#print rows[0][number]
			# IAccessibleChildID 0 is the table it-self, IAccessibleChildID 1 is the list of headers and a first row starts at 2
			childID=rows[0][number]+2
			IAccessibles.append(IAccessible(windowHandle=self.windowHandle,IAccessibleObject=self.IAccessibleObject,IAccessibleChildID=childID,event_windowHandle=self.event_windowHandle,event_objectID=self.event_objectID,event_childID=childID))
		return IAccessibles

class MozillaTableRow(IAccessible):

	def getcolumnText(self,index):
		try:
			if index >=self.parent.columnCount or index <0:
				return None
		except NotImplementedError:
			return None
		try:
			return self.children[index].name
		except IndexError:
			return None

	def getcolumnHeaderText(self,index):
		try:
			listObj=self.parent.firstChild
		except AttributeError:
			return None
		if listObj.role != controlTypes.ROLE_LIST:
			return None
		try:
			return listObj.children[index].name
		except IndexError:
			return None

	def formatColumn(self,index):
		return self.getcolumnText(index)


class ListWithColumnNavigation(NVDAObject):

	def _get_selectionText(self):
		rows=self.selectedRows
		if not rows:
			return None
		names=[]
		for row in rows:
			if row.name:
				names.append(row.name)
		return u"\r".join(names)


class ListItemWithColumnNavigation(NVDAObject):

	colNavItem=0
	lastNumber=0

	def initOverlayClass(self):
		try:
			for n in range(min(self.parent.columnCount,9)):
				self.bindGesture("kb:NVDA+control+%d" % (n+1), "readColumn")
		except NotImplementedError:
			return
		self.bindGesture("kb:control+alt+rightArrow","readNextColumn")
		self.bindGesture("kb:control+alt+leftArrow","readPreviousColumn")
		#self.bindGesture("kb:control+shift+NVDA+downArrow","readCurrentSubitem")
		#self.bindGesture("kb:control+shift+NVDA+upArrow","resetSubitemNumber")

	def script_resetSubitemNumber(self,gesture):
		self.colNavItem=0
		ui.message(_("List navigation: Starting at first column"))
	script_resetSubitemNumber.__doc__=_("Sets the first colum for lists navigation")

	def script_readCurrentSubitem(self,gesture):
		text=self.formatColumn()
		if scriptHandler.getLastScriptRepeatCount()==1 and api.copyToClip(text):
			ui.message(_("%s copied to clipboard")%text)
		else:
			ui.message(text)
	script_readCurrentSubitem.__doc__=_("Reads the current column item when using columns lists navigation. Pressed to times, copies the text to clipboard.")

	def script_readPreviousColumn(self,gesture):
		if self.colNavItem==0:
			self.colNavItem = self.parent.columnCount-1
		else:
			self.colNavItem-=1
		ui.message(self.formatColumn(self.colNavItem))
	script_readPreviousColumn.__doc__=_("Reads the prior column item when using columns lists navigation.")

	def script_readNextColumn(self,gesture):
		if self.colNavItem==self.parent.columnCount-1:
			self.colNavItem = 0
		else:
			self.colNavItem+=1
		ui.message(self.formatColumn(self.colNavItem))
	script_readNextColumn.__doc__=_("Reads the next column item when using columns lists navigation.")

	def script_readColumn(self,gesture):
		try:
			num=int(gesture.keyName[-1])
		except AttributeError:
			num=int(gesture.mainKeyName[-1])
		if num >10 or num >self.parent.columnCount+1:
			return
		text=self.formatColumn(num-1)
		if not text:
			return
		if scriptHandler.getLastScriptRepeatCount()==1 and num ==self.lastNum:
			if api.copyToClip(text):
				ui.message(_("%s copied to clipboard")%text)
		else:
			ui.message(text)
		self.lastNum=num
	script_readColumn.__doc__=_("Presents a listview column corresponding to the keypress from left to right starting at 1")

	def makeTextInfo(self,position):
		if position==textInfos.POSITION_SELECTION:
			position=textInfos.POSITION_ALL
			try:
				text=self.parent.selectionText
				if not text:
					raise NotImplementedError
				self.basicText =text
			except AttributeError:
				raise NotImplementedError
		return super(ListItemWithColumnNavigation,self).makeTextInfo(position)


class GlobalPlugin(globalPluginHandler.GlobalPlugin):
	"""
This is our plugin class, that must be present in order for plugin to work. It must be named exactly as shown above, otherwise the plugin will not be recognized by NVDA.
"""

	def chooseNVDAObjectOverlayClasses(self, obj, clsList):
		if obj.role == controlTypes.ROLE_WINDOW: 
			return
		if obj.appModule.appName == 'egui':
			return
		if obj.appModule.appName == 'utorrent' and obj.windowControlID == 1206 and role == controlTypes.ROLE_LISTITEM:
			return
		windowClass = obj.windowClassName
		if isinstance(obj, IAccessible) and obj.IAccessibleRole == oleacc.ROLE_SYSTEM_LIST and windowClass=="SysListView32":
			clsList.insert(0, ListWithColumnNavigation)
			clsList.insert(0, EnhancedList)
		if isinstance(obj, IAccessible) and obj.IAccessibleRole in (oleacc.ROLE_SYSTEM_TABLE, oleacc.ROLE_SYSTEM_OUTLINE) and windowClass=="MozillaWindowClass":
			clsList.insert(0, ListWithColumnNavigation)
			clsList.insert(0, MozillaTable)
		if isinstance(obj, IAccessible) and obj.IAccessibleRole == oleacc.ROLE_SYSTEM_LISTITEM and windowClass=="SysListView32":
			clsList.insert(0, ListItemWithColumnNavigation)
			clsList.insert(0, EnhancedListItem)
		if isinstance(obj, IAccessible) and obj.IAccessibleRole in (oleacc.ROLE_SYSTEM_ROW, oleacc.ROLE_SYSTEM_OUTLINEITEM) and windowClass=="MozillaWindowClass":
			clsList.insert(0, ListItemWithColumnNavigation)
			clsList.insert(0, MozillaTableRow)
